from __future__ import annotations

from typing import Optional, Dict, Any, Type, TYPE_CHECKING

from taproot.util import PretrainedModelMixin

if TYPE_CHECKING:
    from transformers.models.whisper.tokenization_whisper_fast import WhisperTokenizerFast  # type: ignore[import-untyped,import-not-found,unused-ignore]

__all__ = [
    "WhisperV3Tokenizer",
    "WhisperV1Tokenizer"
]

class WhisperV3Tokenizer(PretrainedModelMixin):
    """
    The tokenizer for Whisper V3 models.
    """
    init_file_urls = {
        "vocab_file": "https://huggingface.co/benjamin-paine/taproot-common/resolve/main/audio-transcription-whisper-tokenizer-v3-vocab.json",
        "merges_file": "https://huggingface.co/benjamin-paine/taproot-common/resolve/main/audio-transcription-whisper-tokenizer-v3-merges.txt",
        "normalizer_file": "https://huggingface.co/benjamin-paine/taproot-common/resolve/main/audio-transcription-whisper-tokenizer-v3-normalizer.json",
        "tokenizer_file": "https://huggingface.co/benjamin-paine/taproot-common/resolve/main/audio-transcription-whisper-tokenizer-v3.json",
    }

    @classmethod
    def get_model_class(cls) -> Type[WhisperTokenizerFast]:
        """
        Get the model class for the Whisper Tokenizer.
        """
        from transformers.models.whisper.tokenization_whisper_fast import WhisperTokenizerFast  # type: ignore[import-untyped,import-not-found,unused-ignore]
        return WhisperTokenizerFast  # type: ignore[no-any-return]

    @classmethod
    def get_default_config(cls) -> Optional[Dict[str, Any]]:
        """
        Get the default configuration for the Whisper  V3 Tokenizer.
        It's a lot.
        """
        from tokenizers import AddedToken # type: ignore[import-untyped,import-not-found,unused-ignore]
        return {
            "add_prefix_space": False,
            "added_tokens_decoder": {
                "50257": AddedToken(
                    content="<|endoftext|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50258": AddedToken(
                    content="<|startoftranscript|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50259": AddedToken(
                    content="<|en|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50260": AddedToken(
                    content="<|zh|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50261": AddedToken(
                    content="<|de|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50262": AddedToken(
                    content="<|es|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50263": AddedToken(
                    content="<|ru|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50264": AddedToken(
                    content="<|ko|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50265": AddedToken(
                    content="<|fr|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50266": AddedToken(
                    content="<|ja|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50267": AddedToken(
                    content="<|pt|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50268": AddedToken(
                    content="<|tr|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50269": AddedToken(
                    content="<|pl|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50270": AddedToken(
                    content="<|ca|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50271": AddedToken(
                    content="<|nl|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50272": AddedToken(
                    content="<|ar|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50273": AddedToken(
                    content="<|sv|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50274": AddedToken(
                    content="<|it|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50275": AddedToken(
                    content="<|id|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50276": AddedToken(
                    content="<|hi|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50277": AddedToken(
                    content="<|fi|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50278": AddedToken(
                    content="<|vi|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50279": AddedToken(
                    content="<|he|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50280": AddedToken(
                    content="<|uk|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50281": AddedToken(
                    content="<|el|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50282": AddedToken(
                    content="<|ms|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50283": AddedToken(
                    content="<|cs|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50284": AddedToken(
                    content="<|ro|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50285": AddedToken(
                    content="<|da|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50286": AddedToken(
                    content="<|hu|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50287": AddedToken(
                    content="<|ta|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50288": AddedToken(
                    content="<|no|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50289": AddedToken(
                    content="<|th|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50290": AddedToken(
                    content="<|ur|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50291": AddedToken(
                    content="<|hr|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50292": AddedToken(
                    content="<|bg|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50293": AddedToken(
                    content="<|lt|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50294": AddedToken(
                    content="<|la|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50295": AddedToken(
                    content="<|mi|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50296": AddedToken(
                    content="<|ml|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50297": AddedToken(
                    content="<|cy|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50298": AddedToken(
                    content="<|sk|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50299": AddedToken(
                    content="<|te|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50300": AddedToken(
                    content="<|fa|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50301": AddedToken(
                    content="<|lv|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50302": AddedToken(
                    content="<|bn|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50303": AddedToken(
                    content="<|sr|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50304": AddedToken(
                    content="<|az|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50305": AddedToken(
                    content="<|sl|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50306": AddedToken(
                    content="<|kn|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50307": AddedToken(
                    content="<|et|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50308": AddedToken(
                    content="<|mk|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50309": AddedToken(
                    content="<|br|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50310": AddedToken(
                    content="<|eu|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50311": AddedToken(
                    content="<|is|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50312": AddedToken(
                    content="<|hy|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50313": AddedToken(
                    content="<|ne|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50314": AddedToken(
                    content="<|mn|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50315": AddedToken(
                    content="<|bs|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50316": AddedToken(
                    content="<|kk|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50317": AddedToken(
                    content="<|sq|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50318": AddedToken(
                    content="<|sw|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50319": AddedToken(
                    content="<|gl|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50320": AddedToken(
                    content="<|mr|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50321": AddedToken(
                    content="<|pa|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50322": AddedToken(
                    content="<|si|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50323": AddedToken(
                    content="<|km|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50324": AddedToken(
                    content="<|sn|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50325": AddedToken(
                    content="<|yo|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50326": AddedToken(
                    content="<|so|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50327": AddedToken(
                    content="<|af|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50328": AddedToken(
                    content="<|oc|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50329": AddedToken(
                    content="<|ka|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50330": AddedToken(
                    content="<|be|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50331": AddedToken(
                    content="<|tg|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50332": AddedToken(
                    content="<|sd|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50333": AddedToken(
                    content="<|gu|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50334": AddedToken(
                    content="<|am|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50335": AddedToken(
                    content="<|yi|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50336": AddedToken(
                    content="<|lo|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50337": AddedToken(
                    content="<|uz|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50338": AddedToken(
                    content="<|fo|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50339": AddedToken(
                    content="<|ht|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50340": AddedToken(
                    content="<|ps|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50341": AddedToken(
                    content="<|tk|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50342": AddedToken(
                    content="<|nn|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50343": AddedToken(
                    content="<|mt|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50344": AddedToken(
                    content="<|sa|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50345": AddedToken(
                    content="<|lb|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50346": AddedToken(
                    content="<|my|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50347": AddedToken(
                    content="<|bo|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50348": AddedToken(
                    content="<|tl|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50349": AddedToken(
                    content="<|mg|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50350": AddedToken(
                    content="<|as|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50351": AddedToken(
                    content="<|tt|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50352": AddedToken(
                    content="<|haw|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50353": AddedToken(
                    content="<|ln|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50354": AddedToken(
                    content="<|ha|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50355": AddedToken(
                    content="<|ba|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50356": AddedToken(
                    content="<|jw|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50357": AddedToken(
                    content="<|su|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50358": AddedToken(
                    content="<|yue|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50359": AddedToken(
                    content="<|translate|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50360": AddedToken(
                    content="<|transcribe|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50361": AddedToken(
                    content="<|startoflm|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50362": AddedToken(
                    content="<|startofprev|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50363": AddedToken(
                    content="<|nospeech|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50364": AddedToken(
                    content="<|notimestamps|>",
                    lstrip=False,
                    normalized=False,
                    rstrip=False,
                    single_word=False,
                    special=True,
                ),
                "50365": AddedToken(
                    content="<|0.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50366": AddedToken(
                    content="<|0.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50367": AddedToken(
                    content="<|0.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50368": AddedToken(
                    content="<|0.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50369": AddedToken(
                    content="<|0.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50370": AddedToken(
                    content="<|0.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50371": AddedToken(
                    content="<|0.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50372": AddedToken(
                    content="<|0.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50373": AddedToken(
                    content="<|0.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50374": AddedToken(
                    content="<|0.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50375": AddedToken(
                    content="<|0.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50376": AddedToken(
                    content="<|0.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50377": AddedToken(
                    content="<|0.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50378": AddedToken(
                    content="<|0.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50379": AddedToken(
                    content="<|0.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50380": AddedToken(
                    content="<|0.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50381": AddedToken(
                    content="<|0.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50382": AddedToken(
                    content="<|0.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50383": AddedToken(
                    content="<|0.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50384": AddedToken(
                    content="<|0.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50385": AddedToken(
                    content="<|0.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50386": AddedToken(
                    content="<|0.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50387": AddedToken(
                    content="<|0.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50388": AddedToken(
                    content="<|0.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50389": AddedToken(
                    content="<|0.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50390": AddedToken(
                    content="<|0.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50391": AddedToken(
                    content="<|0.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50392": AddedToken(
                    content="<|0.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50393": AddedToken(
                    content="<|0.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50394": AddedToken(
                    content="<|0.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50395": AddedToken(
                    content="<|0.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50396": AddedToken(
                    content="<|0.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50397": AddedToken(
                    content="<|0.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50398": AddedToken(
                    content="<|0.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50399": AddedToken(
                    content="<|0.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50400": AddedToken(
                    content="<|0.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50401": AddedToken(
                    content="<|0.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50402": AddedToken(
                    content="<|0.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50403": AddedToken(
                    content="<|0.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50404": AddedToken(
                    content="<|0.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50405": AddedToken(
                    content="<|0.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50406": AddedToken(
                    content="<|0.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50407": AddedToken(
                    content="<|0.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50408": AddedToken(
                    content="<|0.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50409": AddedToken(
                    content="<|0.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50410": AddedToken(
                    content="<|0.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50411": AddedToken(
                    content="<|0.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50412": AddedToken(
                    content="<|0.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50413": AddedToken(
                    content="<|0.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50414": AddedToken(
                    content="<|0.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50415": AddedToken(
                    content="<|1.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50416": AddedToken(
                    content="<|1.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50417": AddedToken(
                    content="<|1.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50418": AddedToken(
                    content="<|1.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50419": AddedToken(
                    content="<|1.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50420": AddedToken(
                    content="<|1.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50421": AddedToken(
                    content="<|1.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50422": AddedToken(
                    content="<|1.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50423": AddedToken(
                    content="<|1.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50424": AddedToken(
                    content="<|1.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50425": AddedToken(
                    content="<|1.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50426": AddedToken(
                    content="<|1.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50427": AddedToken(
                    content="<|1.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50428": AddedToken(
                    content="<|1.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50429": AddedToken(
                    content="<|1.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50430": AddedToken(
                    content="<|1.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50431": AddedToken(
                    content="<|1.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50432": AddedToken(
                    content="<|1.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50433": AddedToken(
                    content="<|1.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50434": AddedToken(
                    content="<|1.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50435": AddedToken(
                    content="<|1.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50436": AddedToken(
                    content="<|1.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50437": AddedToken(
                    content="<|1.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50438": AddedToken(
                    content="<|1.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50439": AddedToken(
                    content="<|1.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50440": AddedToken(
                    content="<|1.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50441": AddedToken(
                    content="<|1.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50442": AddedToken(
                    content="<|1.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50443": AddedToken(
                    content="<|1.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50444": AddedToken(
                    content="<|1.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50445": AddedToken(
                    content="<|1.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50446": AddedToken(
                    content="<|1.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50447": AddedToken(
                    content="<|1.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50448": AddedToken(
                    content="<|1.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50449": AddedToken(
                    content="<|1.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50450": AddedToken(
                    content="<|1.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50451": AddedToken(
                    content="<|1.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50452": AddedToken(
                    content="<|1.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50453": AddedToken(
                    content="<|1.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50454": AddedToken(
                    content="<|1.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50455": AddedToken(
                    content="<|1.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50456": AddedToken(
                    content="<|1.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50457": AddedToken(
                    content="<|1.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50458": AddedToken(
                    content="<|1.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50459": AddedToken(
                    content="<|1.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50460": AddedToken(
                    content="<|1.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50461": AddedToken(
                    content="<|1.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50462": AddedToken(
                    content="<|1.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50463": AddedToken(
                    content="<|1.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50464": AddedToken(
                    content="<|1.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50465": AddedToken(
                    content="<|2.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50466": AddedToken(
                    content="<|2.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50467": AddedToken(
                    content="<|2.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50468": AddedToken(
                    content="<|2.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50469": AddedToken(
                    content="<|2.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50470": AddedToken(
                    content="<|2.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50471": AddedToken(
                    content="<|2.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50472": AddedToken(
                    content="<|2.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50473": AddedToken(
                    content="<|2.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50474": AddedToken(
                    content="<|2.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50475": AddedToken(
                    content="<|2.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50476": AddedToken(
                    content="<|2.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50477": AddedToken(
                    content="<|2.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50478": AddedToken(
                    content="<|2.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50479": AddedToken(
                    content="<|2.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50480": AddedToken(
                    content="<|2.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50481": AddedToken(
                    content="<|2.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50482": AddedToken(
                    content="<|2.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50483": AddedToken(
                    content="<|2.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50484": AddedToken(
                    content="<|2.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50485": AddedToken(
                    content="<|2.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50486": AddedToken(
                    content="<|2.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50487": AddedToken(
                    content="<|2.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50488": AddedToken(
                    content="<|2.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50489": AddedToken(
                    content="<|2.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50490": AddedToken(
                    content="<|2.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50491": AddedToken(
                    content="<|2.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50492": AddedToken(
                    content="<|2.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50493": AddedToken(
                    content="<|2.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50494": AddedToken(
                    content="<|2.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50495": AddedToken(
                    content="<|2.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50496": AddedToken(
                    content="<|2.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50497": AddedToken(
                    content="<|2.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50498": AddedToken(
                    content="<|2.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50499": AddedToken(
                    content="<|2.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50500": AddedToken(
                    content="<|2.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50501": AddedToken(
                    content="<|2.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50502": AddedToken(
                    content="<|2.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50503": AddedToken(
                    content="<|2.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50504": AddedToken(
                    content="<|2.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50505": AddedToken(
                    content="<|2.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50506": AddedToken(
                    content="<|2.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50507": AddedToken(
                    content="<|2.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50508": AddedToken(
                    content="<|2.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50509": AddedToken(
                    content="<|2.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50510": AddedToken(
                    content="<|2.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50511": AddedToken(
                    content="<|2.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50512": AddedToken(
                    content="<|2.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50513": AddedToken(
                    content="<|2.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50514": AddedToken(
                    content="<|2.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50515": AddedToken(
                    content="<|3.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50516": AddedToken(
                    content="<|3.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50517": AddedToken(
                    content="<|3.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50518": AddedToken(
                    content="<|3.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50519": AddedToken(
                    content="<|3.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50520": AddedToken(
                    content="<|3.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50521": AddedToken(
                    content="<|3.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50522": AddedToken(
                    content="<|3.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50523": AddedToken(
                    content="<|3.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50524": AddedToken(
                    content="<|3.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50525": AddedToken(
                    content="<|3.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50526": AddedToken(
                    content="<|3.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50527": AddedToken(
                    content="<|3.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50528": AddedToken(
                    content="<|3.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50529": AddedToken(
                    content="<|3.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50530": AddedToken(
                    content="<|3.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50531": AddedToken(
                    content="<|3.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50532": AddedToken(
                    content="<|3.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50533": AddedToken(
                    content="<|3.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50534": AddedToken(
                    content="<|3.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50535": AddedToken(
                    content="<|3.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50536": AddedToken(
                    content="<|3.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50537": AddedToken(
                    content="<|3.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50538": AddedToken(
                    content="<|3.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50539": AddedToken(
                    content="<|3.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50540": AddedToken(
                    content="<|3.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50541": AddedToken(
                    content="<|3.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50542": AddedToken(
                    content="<|3.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50543": AddedToken(
                    content="<|3.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50544": AddedToken(
                    content="<|3.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50545": AddedToken(
                    content="<|3.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50546": AddedToken(
                    content="<|3.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50547": AddedToken(
                    content="<|3.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50548": AddedToken(
                    content="<|3.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50549": AddedToken(
                    content="<|3.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50550": AddedToken(
                    content="<|3.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50551": AddedToken(
                    content="<|3.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50552": AddedToken(
                    content="<|3.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50553": AddedToken(
                    content="<|3.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50554": AddedToken(
                    content="<|3.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50555": AddedToken(
                    content="<|3.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50556": AddedToken(
                    content="<|3.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50557": AddedToken(
                    content="<|3.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50558": AddedToken(
                    content="<|3.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50559": AddedToken(
                    content="<|3.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50560": AddedToken(
                    content="<|3.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50561": AddedToken(
                    content="<|3.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50562": AddedToken(
                    content="<|3.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50563": AddedToken(
                    content="<|3.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50564": AddedToken(
                    content="<|3.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50565": AddedToken(
                    content="<|4.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50566": AddedToken(
                    content="<|4.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50567": AddedToken(
                    content="<|4.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50568": AddedToken(
                    content="<|4.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50569": AddedToken(
                    content="<|4.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50570": AddedToken(
                    content="<|4.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50571": AddedToken(
                    content="<|4.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50572": AddedToken(
                    content="<|4.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50573": AddedToken(
                    content="<|4.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50574": AddedToken(
                    content="<|4.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50575": AddedToken(
                    content="<|4.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50576": AddedToken(
                    content="<|4.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50577": AddedToken(
                    content="<|4.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50578": AddedToken(
                    content="<|4.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50579": AddedToken(
                    content="<|4.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50580": AddedToken(
                    content="<|4.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50581": AddedToken(
                    content="<|4.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50582": AddedToken(
                    content="<|4.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50583": AddedToken(
                    content="<|4.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50584": AddedToken(
                    content="<|4.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50585": AddedToken(
                    content="<|4.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50586": AddedToken(
                    content="<|4.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50587": AddedToken(
                    content="<|4.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50588": AddedToken(
                    content="<|4.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50589": AddedToken(
                    content="<|4.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50590": AddedToken(
                    content="<|4.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50591": AddedToken(
                    content="<|4.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50592": AddedToken(
                    content="<|4.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50593": AddedToken(
                    content="<|4.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50594": AddedToken(
                    content="<|4.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50595": AddedToken(
                    content="<|4.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50596": AddedToken(
                    content="<|4.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50597": AddedToken(
                    content="<|4.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50598": AddedToken(
                    content="<|4.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50599": AddedToken(
                    content="<|4.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50600": AddedToken(
                    content="<|4.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50601": AddedToken(
                    content="<|4.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50602": AddedToken(
                    content="<|4.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50603": AddedToken(
                    content="<|4.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50604": AddedToken(
                    content="<|4.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50605": AddedToken(
                    content="<|4.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50606": AddedToken(
                    content="<|4.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50607": AddedToken(
                    content="<|4.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50608": AddedToken(
                    content="<|4.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50609": AddedToken(
                    content="<|4.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50610": AddedToken(
                    content="<|4.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50611": AddedToken(
                    content="<|4.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50612": AddedToken(
                    content="<|4.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50613": AddedToken(
                    content="<|4.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50614": AddedToken(
                    content="<|4.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50615": AddedToken(
                    content="<|5.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50616": AddedToken(
                    content="<|5.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50617": AddedToken(
                    content="<|5.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50618": AddedToken(
                    content="<|5.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50619": AddedToken(
                    content="<|5.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50620": AddedToken(
                    content="<|5.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50621": AddedToken(
                    content="<|5.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50622": AddedToken(
                    content="<|5.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50623": AddedToken(
                    content="<|5.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50624": AddedToken(
                    content="<|5.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50625": AddedToken(
                    content="<|5.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50626": AddedToken(
                    content="<|5.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50627": AddedToken(
                    content="<|5.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50628": AddedToken(
                    content="<|5.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50629": AddedToken(
                    content="<|5.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50630": AddedToken(
                    content="<|5.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50631": AddedToken(
                    content="<|5.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50632": AddedToken(
                    content="<|5.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50633": AddedToken(
                    content="<|5.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50634": AddedToken(
                    content="<|5.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50635": AddedToken(
                    content="<|5.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50636": AddedToken(
                    content="<|5.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50637": AddedToken(
                    content="<|5.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50638": AddedToken(
                    content="<|5.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50639": AddedToken(
                    content="<|5.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50640": AddedToken(
                    content="<|5.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50641": AddedToken(
                    content="<|5.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50642": AddedToken(
                    content="<|5.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50643": AddedToken(
                    content="<|5.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50644": AddedToken(
                    content="<|5.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50645": AddedToken(
                    content="<|5.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50646": AddedToken(
                    content="<|5.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50647": AddedToken(
                    content="<|5.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50648": AddedToken(
                    content="<|5.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50649": AddedToken(
                    content="<|5.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50650": AddedToken(
                    content="<|5.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50651": AddedToken(
                    content="<|5.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50652": AddedToken(
                    content="<|5.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50653": AddedToken(
                    content="<|5.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50654": AddedToken(
                    content="<|5.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50655": AddedToken(
                    content="<|5.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50656": AddedToken(
                    content="<|5.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50657": AddedToken(
                    content="<|5.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50658": AddedToken(
                    content="<|5.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50659": AddedToken(
                    content="<|5.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50660": AddedToken(
                    content="<|5.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50661": AddedToken(
                    content="<|5.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50662": AddedToken(
                    content="<|5.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50663": AddedToken(
                    content="<|5.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50664": AddedToken(
                    content="<|5.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50665": AddedToken(
                    content="<|6.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50666": AddedToken(
                    content="<|6.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50667": AddedToken(
                    content="<|6.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50668": AddedToken(
                    content="<|6.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50669": AddedToken(
                    content="<|6.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50670": AddedToken(
                    content="<|6.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50671": AddedToken(
                    content="<|6.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50672": AddedToken(
                    content="<|6.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50673": AddedToken(
                    content="<|6.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50674": AddedToken(
                    content="<|6.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50675": AddedToken(
                    content="<|6.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50676": AddedToken(
                    content="<|6.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50677": AddedToken(
                    content="<|6.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50678": AddedToken(
                    content="<|6.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50679": AddedToken(
                    content="<|6.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50680": AddedToken(
                    content="<|6.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50681": AddedToken(
                    content="<|6.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50682": AddedToken(
                    content="<|6.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50683": AddedToken(
                    content="<|6.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50684": AddedToken(
                    content="<|6.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50685": AddedToken(
                    content="<|6.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50686": AddedToken(
                    content="<|6.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50687": AddedToken(
                    content="<|6.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50688": AddedToken(
                    content="<|6.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50689": AddedToken(
                    content="<|6.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50690": AddedToken(
                    content="<|6.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50691": AddedToken(
                    content="<|6.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50692": AddedToken(
                    content="<|6.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50693": AddedToken(
                    content="<|6.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50694": AddedToken(
                    content="<|6.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50695": AddedToken(
                    content="<|6.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50696": AddedToken(
                    content="<|6.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50697": AddedToken(
                    content="<|6.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50698": AddedToken(
                    content="<|6.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50699": AddedToken(
                    content="<|6.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50700": AddedToken(
                    content="<|6.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50701": AddedToken(
                    content="<|6.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50702": AddedToken(
                    content="<|6.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50703": AddedToken(
                    content="<|6.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50704": AddedToken(
                    content="<|6.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50705": AddedToken(
                    content="<|6.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50706": AddedToken(
                    content="<|6.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50707": AddedToken(
                    content="<|6.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50708": AddedToken(
                    content="<|6.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50709": AddedToken(
                    content="<|6.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50710": AddedToken(
                    content="<|6.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50711": AddedToken(
                    content="<|6.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50712": AddedToken(
                    content="<|6.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50713": AddedToken(
                    content="<|6.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50714": AddedToken(
                    content="<|6.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50715": AddedToken(
                    content="<|7.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50716": AddedToken(
                    content="<|7.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50717": AddedToken(
                    content="<|7.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50718": AddedToken(
                    content="<|7.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50719": AddedToken(
                    content="<|7.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50720": AddedToken(
                    content="<|7.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50721": AddedToken(
                    content="<|7.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50722": AddedToken(
                    content="<|7.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50723": AddedToken(
                    content="<|7.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50724": AddedToken(
                    content="<|7.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50725": AddedToken(
                    content="<|7.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50726": AddedToken(
                    content="<|7.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50727": AddedToken(
                    content="<|7.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50728": AddedToken(
                    content="<|7.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50729": AddedToken(
                    content="<|7.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50730": AddedToken(
                    content="<|7.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50731": AddedToken(
                    content="<|7.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50732": AddedToken(
                    content="<|7.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50733": AddedToken(
                    content="<|7.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50734": AddedToken(
                    content="<|7.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50735": AddedToken(
                    content="<|7.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50736": AddedToken(
                    content="<|7.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50737": AddedToken(
                    content="<|7.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50738": AddedToken(
                    content="<|7.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50739": AddedToken(
                    content="<|7.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50740": AddedToken(
                    content="<|7.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50741": AddedToken(
                    content="<|7.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50742": AddedToken(
                    content="<|7.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50743": AddedToken(
                    content="<|7.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50744": AddedToken(
                    content="<|7.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50745": AddedToken(
                    content="<|7.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50746": AddedToken(
                    content="<|7.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50747": AddedToken(
                    content="<|7.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50748": AddedToken(
                    content="<|7.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50749": AddedToken(
                    content="<|7.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50750": AddedToken(
                    content="<|7.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50751": AddedToken(
                    content="<|7.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50752": AddedToken(
                    content="<|7.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50753": AddedToken(
                    content="<|7.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50754": AddedToken(
                    content="<|7.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50755": AddedToken(
                    content="<|7.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50756": AddedToken(
                    content="<|7.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50757": AddedToken(
                    content="<|7.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50758": AddedToken(
                    content="<|7.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50759": AddedToken(
                    content="<|7.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50760": AddedToken(
                    content="<|7.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50761": AddedToken(
                    content="<|7.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50762": AddedToken(
                    content="<|7.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50763": AddedToken(
                    content="<|7.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50764": AddedToken(
                    content="<|7.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50765": AddedToken(
                    content="<|8.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50766": AddedToken(
                    content="<|8.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50767": AddedToken(
                    content="<|8.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50768": AddedToken(
                    content="<|8.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50769": AddedToken(
                    content="<|8.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50770": AddedToken(
                    content="<|8.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50771": AddedToken(
                    content="<|8.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50772": AddedToken(
                    content="<|8.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50773": AddedToken(
                    content="<|8.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50774": AddedToken(
                    content="<|8.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50775": AddedToken(
                    content="<|8.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50776": AddedToken(
                    content="<|8.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50777": AddedToken(
                    content="<|8.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50778": AddedToken(
                    content="<|8.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50779": AddedToken(
                    content="<|8.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50780": AddedToken(
                    content="<|8.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50781": AddedToken(
                    content="<|8.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50782": AddedToken(
                    content="<|8.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50783": AddedToken(
                    content="<|8.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50784": AddedToken(
                    content="<|8.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50785": AddedToken(
                    content="<|8.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50786": AddedToken(
                    content="<|8.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50787": AddedToken(
                    content="<|8.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50788": AddedToken(
                    content="<|8.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50789": AddedToken(
                    content="<|8.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50790": AddedToken(
                    content="<|8.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50791": AddedToken(
                    content="<|8.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50792": AddedToken(
                    content="<|8.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50793": AddedToken(
                    content="<|8.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50794": AddedToken(
                    content="<|8.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50795": AddedToken(
                    content="<|8.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50796": AddedToken(
                    content="<|8.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50797": AddedToken(
                    content="<|8.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50798": AddedToken(
                    content="<|8.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50799": AddedToken(
                    content="<|8.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50800": AddedToken(
                    content="<|8.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50801": AddedToken(
                    content="<|8.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50802": AddedToken(
                    content="<|8.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50803": AddedToken(
                    content="<|8.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50804": AddedToken(
                    content="<|8.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50805": AddedToken(
                    content="<|8.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50806": AddedToken(
                    content="<|8.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50807": AddedToken(
                    content="<|8.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50808": AddedToken(
                    content="<|8.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50809": AddedToken(
                    content="<|8.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50810": AddedToken(
                    content="<|8.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50811": AddedToken(
                    content="<|8.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50812": AddedToken(
                    content="<|8.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50813": AddedToken(
                    content="<|8.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50814": AddedToken(
                    content="<|8.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50815": AddedToken(
                    content="<|9.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50816": AddedToken(
                    content="<|9.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50817": AddedToken(
                    content="<|9.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50818": AddedToken(
                    content="<|9.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50819": AddedToken(
                    content="<|9.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50820": AddedToken(
                    content="<|9.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50821": AddedToken(
                    content="<|9.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50822": AddedToken(
                    content="<|9.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50823": AddedToken(
                    content="<|9.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50824": AddedToken(
                    content="<|9.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50825": AddedToken(
                    content="<|9.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50826": AddedToken(
                    content="<|9.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50827": AddedToken(
                    content="<|9.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50828": AddedToken(
                    content="<|9.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50829": AddedToken(
                    content="<|9.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50830": AddedToken(
                    content="<|9.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50831": AddedToken(
                    content="<|9.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50832": AddedToken(
                    content="<|9.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50833": AddedToken(
                    content="<|9.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50834": AddedToken(
                    content="<|9.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50835": AddedToken(
                    content="<|9.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50836": AddedToken(
                    content="<|9.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50837": AddedToken(
                    content="<|9.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50838": AddedToken(
                    content="<|9.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50839": AddedToken(
                    content="<|9.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50840": AddedToken(
                    content="<|9.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50841": AddedToken(
                    content="<|9.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50842": AddedToken(
                    content="<|9.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50843": AddedToken(
                    content="<|9.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50844": AddedToken(
                    content="<|9.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50845": AddedToken(
                    content="<|9.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50846": AddedToken(
                    content="<|9.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50847": AddedToken(
                    content="<|9.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50848": AddedToken(
                    content="<|9.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50849": AddedToken(
                    content="<|9.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50850": AddedToken(
                    content="<|9.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50851": AddedToken(
                    content="<|9.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50852": AddedToken(
                    content="<|9.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50853": AddedToken(
                    content="<|9.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50854": AddedToken(
                    content="<|9.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50855": AddedToken(
                    content="<|9.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50856": AddedToken(
                    content="<|9.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50857": AddedToken(
                    content="<|9.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50858": AddedToken(
                    content="<|9.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50859": AddedToken(
                    content="<|9.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50860": AddedToken(
                    content="<|9.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50861": AddedToken(
                    content="<|9.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50862": AddedToken(
                    content="<|9.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50863": AddedToken(
                    content="<|9.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50864": AddedToken(
                    content="<|9.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50865": AddedToken(
                    content="<|10.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50866": AddedToken(
                    content="<|10.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50867": AddedToken(
                    content="<|10.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50868": AddedToken(
                    content="<|10.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50869": AddedToken(
                    content="<|10.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50870": AddedToken(
                    content="<|10.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50871": AddedToken(
                    content="<|10.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50872": AddedToken(
                    content="<|10.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50873": AddedToken(
                    content="<|10.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50874": AddedToken(
                    content="<|10.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50875": AddedToken(
                    content="<|10.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50876": AddedToken(
                    content="<|10.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50877": AddedToken(
                    content="<|10.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50878": AddedToken(
                    content="<|10.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50879": AddedToken(
                    content="<|10.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50880": AddedToken(
                    content="<|10.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50881": AddedToken(
                    content="<|10.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50882": AddedToken(
                    content="<|10.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50883": AddedToken(
                    content="<|10.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50884": AddedToken(
                    content="<|10.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50885": AddedToken(
                    content="<|10.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50886": AddedToken(
                    content="<|10.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50887": AddedToken(
                    content="<|10.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50888": AddedToken(
                    content="<|10.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50889": AddedToken(
                    content="<|10.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50890": AddedToken(
                    content="<|10.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50891": AddedToken(
                    content="<|10.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50892": AddedToken(
                    content="<|10.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50893": AddedToken(
                    content="<|10.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50894": AddedToken(
                    content="<|10.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50895": AddedToken(
                    content="<|10.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50896": AddedToken(
                    content="<|10.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50897": AddedToken(
                    content="<|10.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50898": AddedToken(
                    content="<|10.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50899": AddedToken(
                    content="<|10.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50900": AddedToken(
                    content="<|10.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50901": AddedToken(
                    content="<|10.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50902": AddedToken(
                    content="<|10.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50903": AddedToken(
                    content="<|10.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50904": AddedToken(
                    content="<|10.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50905": AddedToken(
                    content="<|10.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50906": AddedToken(
                    content="<|10.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50907": AddedToken(
                    content="<|10.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50908": AddedToken(
                    content="<|10.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50909": AddedToken(
                    content="<|10.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50910": AddedToken(
                    content="<|10.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50911": AddedToken(
                    content="<|10.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50912": AddedToken(
                    content="<|10.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50913": AddedToken(
                    content="<|10.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50914": AddedToken(
                    content="<|10.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50915": AddedToken(
                    content="<|11.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50916": AddedToken(
                    content="<|11.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50917": AddedToken(
                    content="<|11.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50918": AddedToken(
                    content="<|11.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50919": AddedToken(
                    content="<|11.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50920": AddedToken(
                    content="<|11.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50921": AddedToken(
                    content="<|11.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50922": AddedToken(
                    content="<|11.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50923": AddedToken(
                    content="<|11.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50924": AddedToken(
                    content="<|11.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50925": AddedToken(
                    content="<|11.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50926": AddedToken(
                    content="<|11.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50927": AddedToken(
                    content="<|11.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50928": AddedToken(
                    content="<|11.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50929": AddedToken(
                    content="<|11.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50930": AddedToken(
                    content="<|11.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50931": AddedToken(
                    content="<|11.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50932": AddedToken(
                    content="<|11.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50933": AddedToken(
                    content="<|11.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50934": AddedToken(
                    content="<|11.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50935": AddedToken(
                    content="<|11.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50936": AddedToken(
                    content="<|11.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50937": AddedToken(
                    content="<|11.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50938": AddedToken(
                    content="<|11.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50939": AddedToken(
                    content="<|11.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50940": AddedToken(
                    content="<|11.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50941": AddedToken(
                    content="<|11.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50942": AddedToken(
                    content="<|11.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50943": AddedToken(
                    content="<|11.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50944": AddedToken(
                    content="<|11.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50945": AddedToken(
                    content="<|11.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50946": AddedToken(
                    content="<|11.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50947": AddedToken(
                    content="<|11.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50948": AddedToken(
                    content="<|11.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50949": AddedToken(
                    content="<|11.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50950": AddedToken(
                    content="<|11.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50951": AddedToken(
                    content="<|11.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50952": AddedToken(
                    content="<|11.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50953": AddedToken(
                    content="<|11.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50954": AddedToken(
                    content="<|11.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50955": AddedToken(
                    content="<|11.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50956": AddedToken(
                    content="<|11.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50957": AddedToken(
                    content="<|11.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50958": AddedToken(
                    content="<|11.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50959": AddedToken(
                    content="<|11.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50960": AddedToken(
                    content="<|11.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50961": AddedToken(
                    content="<|11.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50962": AddedToken(
                    content="<|11.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50963": AddedToken(
                    content="<|11.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50964": AddedToken(
                    content="<|11.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50965": AddedToken(
                    content="<|12.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50966": AddedToken(
                    content="<|12.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50967": AddedToken(
                    content="<|12.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50968": AddedToken(
                    content="<|12.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50969": AddedToken(
                    content="<|12.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50970": AddedToken(
                    content="<|12.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50971": AddedToken(
                    content="<|12.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50972": AddedToken(
                    content="<|12.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50973": AddedToken(
                    content="<|12.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50974": AddedToken(
                    content="<|12.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50975": AddedToken(
                    content="<|12.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50976": AddedToken(
                    content="<|12.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50977": AddedToken(
                    content="<|12.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50978": AddedToken(
                    content="<|12.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50979": AddedToken(
                    content="<|12.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50980": AddedToken(
                    content="<|12.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50981": AddedToken(
                    content="<|12.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50982": AddedToken(
                    content="<|12.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50983": AddedToken(
                    content="<|12.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50984": AddedToken(
                    content="<|12.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50985": AddedToken(
                    content="<|12.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50986": AddedToken(
                    content="<|12.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50987": AddedToken(
                    content="<|12.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50988": AddedToken(
                    content="<|12.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50989": AddedToken(
                    content="<|12.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50990": AddedToken(
                    content="<|12.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50991": AddedToken(
                    content="<|12.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50992": AddedToken(
                    content="<|12.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50993": AddedToken(
                    content="<|12.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50994": AddedToken(
                    content="<|12.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50995": AddedToken(
                    content="<|12.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50996": AddedToken(
                    content="<|12.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50997": AddedToken(
                    content="<|12.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50998": AddedToken(
                    content="<|12.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "50999": AddedToken(
                    content="<|12.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51000": AddedToken(
                    content="<|12.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51001": AddedToken(
                    content="<|12.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51002": AddedToken(
                    content="<|12.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51003": AddedToken(
                    content="<|12.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51004": AddedToken(
                    content="<|12.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51005": AddedToken(
                    content="<|12.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51006": AddedToken(
                    content="<|12.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51007": AddedToken(
                    content="<|12.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51008": AddedToken(
                    content="<|12.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51009": AddedToken(
                    content="<|12.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51010": AddedToken(
                    content="<|12.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51011": AddedToken(
                    content="<|12.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51012": AddedToken(
                    content="<|12.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51013": AddedToken(
                    content="<|12.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51014": AddedToken(
                    content="<|12.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51015": AddedToken(
                    content="<|13.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51016": AddedToken(
                    content="<|13.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51017": AddedToken(
                    content="<|13.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51018": AddedToken(
                    content="<|13.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51019": AddedToken(
                    content="<|13.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51020": AddedToken(
                    content="<|13.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51021": AddedToken(
                    content="<|13.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51022": AddedToken(
                    content="<|13.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51023": AddedToken(
                    content="<|13.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51024": AddedToken(
                    content="<|13.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51025": AddedToken(
                    content="<|13.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51026": AddedToken(
                    content="<|13.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51027": AddedToken(
                    content="<|13.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51028": AddedToken(
                    content="<|13.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51029": AddedToken(
                    content="<|13.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51030": AddedToken(
                    content="<|13.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51031": AddedToken(
                    content="<|13.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51032": AddedToken(
                    content="<|13.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51033": AddedToken(
                    content="<|13.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51034": AddedToken(
                    content="<|13.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51035": AddedToken(
                    content="<|13.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51036": AddedToken(
                    content="<|13.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51037": AddedToken(
                    content="<|13.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51038": AddedToken(
                    content="<|13.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51039": AddedToken(
                    content="<|13.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51040": AddedToken(
                    content="<|13.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51041": AddedToken(
                    content="<|13.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51042": AddedToken(
                    content="<|13.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51043": AddedToken(
                    content="<|13.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51044": AddedToken(
                    content="<|13.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51045": AddedToken(
                    content="<|13.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51046": AddedToken(
                    content="<|13.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51047": AddedToken(
                    content="<|13.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51048": AddedToken(
                    content="<|13.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51049": AddedToken(
                    content="<|13.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51050": AddedToken(
                    content="<|13.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51051": AddedToken(
                    content="<|13.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51052": AddedToken(
                    content="<|13.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51053": AddedToken(
                    content="<|13.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51054": AddedToken(
                    content="<|13.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51055": AddedToken(
                    content="<|13.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51056": AddedToken(
                    content="<|13.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51057": AddedToken(
                    content="<|13.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51058": AddedToken(
                    content="<|13.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51059": AddedToken(
                    content="<|13.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51060": AddedToken(
                    content="<|13.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51061": AddedToken(
                    content="<|13.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51062": AddedToken(
                    content="<|13.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51063": AddedToken(
                    content="<|13.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51064": AddedToken(
                    content="<|13.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51065": AddedToken(
                    content="<|14.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51066": AddedToken(
                    content="<|14.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51067": AddedToken(
                    content="<|14.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51068": AddedToken(
                    content="<|14.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51069": AddedToken(
                    content="<|14.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51070": AddedToken(
                    content="<|14.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51071": AddedToken(
                    content="<|14.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51072": AddedToken(
                    content="<|14.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51073": AddedToken(
                    content="<|14.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51074": AddedToken(
                    content="<|14.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51075": AddedToken(
                    content="<|14.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51076": AddedToken(
                    content="<|14.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51077": AddedToken(
                    content="<|14.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51078": AddedToken(
                    content="<|14.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51079": AddedToken(
                    content="<|14.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51080": AddedToken(
                    content="<|14.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51081": AddedToken(
                    content="<|14.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51082": AddedToken(
                    content="<|14.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51083": AddedToken(
                    content="<|14.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51084": AddedToken(
                    content="<|14.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51085": AddedToken(
                    content="<|14.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51086": AddedToken(
                    content="<|14.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51087": AddedToken(
                    content="<|14.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51088": AddedToken(
                    content="<|14.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51089": AddedToken(
                    content="<|14.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51090": AddedToken(
                    content="<|14.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51091": AddedToken(
                    content="<|14.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51092": AddedToken(
                    content="<|14.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51093": AddedToken(
                    content="<|14.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51094": AddedToken(
                    content="<|14.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51095": AddedToken(
                    content="<|14.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51096": AddedToken(
                    content="<|14.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51097": AddedToken(
                    content="<|14.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51098": AddedToken(
                    content="<|14.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51099": AddedToken(
                    content="<|14.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51100": AddedToken(
                    content="<|14.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51101": AddedToken(
                    content="<|14.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51102": AddedToken(
                    content="<|14.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51103": AddedToken(
                    content="<|14.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51104": AddedToken(
                    content="<|14.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51105": AddedToken(
                    content="<|14.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51106": AddedToken(
                    content="<|14.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51107": AddedToken(
                    content="<|14.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51108": AddedToken(
                    content="<|14.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51109": AddedToken(
                    content="<|14.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51110": AddedToken(
                    content="<|14.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51111": AddedToken(
                    content="<|14.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51112": AddedToken(
                    content="<|14.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51113": AddedToken(
                    content="<|14.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51114": AddedToken(
                    content="<|14.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51115": AddedToken(
                    content="<|15.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51116": AddedToken(
                    content="<|15.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51117": AddedToken(
                    content="<|15.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51118": AddedToken(
                    content="<|15.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51119": AddedToken(
                    content="<|15.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51120": AddedToken(
                    content="<|15.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51121": AddedToken(
                    content="<|15.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51122": AddedToken(
                    content="<|15.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51123": AddedToken(
                    content="<|15.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51124": AddedToken(
                    content="<|15.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51125": AddedToken(
                    content="<|15.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51126": AddedToken(
                    content="<|15.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51127": AddedToken(
                    content="<|15.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51128": AddedToken(
                    content="<|15.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51129": AddedToken(
                    content="<|15.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51130": AddedToken(
                    content="<|15.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51131": AddedToken(
                    content="<|15.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51132": AddedToken(
                    content="<|15.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51133": AddedToken(
                    content="<|15.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51134": AddedToken(
                    content="<|15.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51135": AddedToken(
                    content="<|15.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51136": AddedToken(
                    content="<|15.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51137": AddedToken(
                    content="<|15.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51138": AddedToken(
                    content="<|15.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51139": AddedToken(
                    content="<|15.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51140": AddedToken(
                    content="<|15.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51141": AddedToken(
                    content="<|15.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51142": AddedToken(
                    content="<|15.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51143": AddedToken(
                    content="<|15.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51144": AddedToken(
                    content="<|15.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51145": AddedToken(
                    content="<|15.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51146": AddedToken(
                    content="<|15.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51147": AddedToken(
                    content="<|15.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51148": AddedToken(
                    content="<|15.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51149": AddedToken(
                    content="<|15.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51150": AddedToken(
                    content="<|15.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51151": AddedToken(
                    content="<|15.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51152": AddedToken(
                    content="<|15.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51153": AddedToken(
                    content="<|15.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51154": AddedToken(
                    content="<|15.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51155": AddedToken(
                    content="<|15.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51156": AddedToken(
                    content="<|15.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51157": AddedToken(
                    content="<|15.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51158": AddedToken(
                    content="<|15.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51159": AddedToken(
                    content="<|15.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51160": AddedToken(
                    content="<|15.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51161": AddedToken(
                    content="<|15.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51162": AddedToken(
                    content="<|15.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51163": AddedToken(
                    content="<|15.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51164": AddedToken(
                    content="<|15.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51165": AddedToken(
                    content="<|16.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51166": AddedToken(
                    content="<|16.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51167": AddedToken(
                    content="<|16.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51168": AddedToken(
                    content="<|16.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51169": AddedToken(
                    content="<|16.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51170": AddedToken(
                    content="<|16.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51171": AddedToken(
                    content="<|16.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51172": AddedToken(
                    content="<|16.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51173": AddedToken(
                    content="<|16.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51174": AddedToken(
                    content="<|16.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51175": AddedToken(
                    content="<|16.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51176": AddedToken(
                    content="<|16.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51177": AddedToken(
                    content="<|16.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51178": AddedToken(
                    content="<|16.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51179": AddedToken(
                    content="<|16.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51180": AddedToken(
                    content="<|16.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51181": AddedToken(
                    content="<|16.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51182": AddedToken(
                    content="<|16.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51183": AddedToken(
                    content="<|16.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51184": AddedToken(
                    content="<|16.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51185": AddedToken(
                    content="<|16.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51186": AddedToken(
                    content="<|16.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51187": AddedToken(
                    content="<|16.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51188": AddedToken(
                    content="<|16.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51189": AddedToken(
                    content="<|16.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51190": AddedToken(
                    content="<|16.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51191": AddedToken(
                    content="<|16.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51192": AddedToken(
                    content="<|16.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51193": AddedToken(
                    content="<|16.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51194": AddedToken(
                    content="<|16.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51195": AddedToken(
                    content="<|16.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51196": AddedToken(
                    content="<|16.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51197": AddedToken(
                    content="<|16.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51198": AddedToken(
                    content="<|16.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51199": AddedToken(
                    content="<|16.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51200": AddedToken(
                    content="<|16.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51201": AddedToken(
                    content="<|16.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51202": AddedToken(
                    content="<|16.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51203": AddedToken(
                    content="<|16.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51204": AddedToken(
                    content="<|16.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51205": AddedToken(
                    content="<|16.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51206": AddedToken(
                    content="<|16.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51207": AddedToken(
                    content="<|16.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51208": AddedToken(
                    content="<|16.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51209": AddedToken(
                    content="<|16.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51210": AddedToken(
                    content="<|16.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51211": AddedToken(
                    content="<|16.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51212": AddedToken(
                    content="<|16.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51213": AddedToken(
                    content="<|16.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51214": AddedToken(
                    content="<|16.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51215": AddedToken(
                    content="<|17.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51216": AddedToken(
                    content="<|17.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51217": AddedToken(
                    content="<|17.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51218": AddedToken(
                    content="<|17.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51219": AddedToken(
                    content="<|17.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51220": AddedToken(
                    content="<|17.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51221": AddedToken(
                    content="<|17.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51222": AddedToken(
                    content="<|17.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51223": AddedToken(
                    content="<|17.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51224": AddedToken(
                    content="<|17.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51225": AddedToken(
                    content="<|17.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51226": AddedToken(
                    content="<|17.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51227": AddedToken(
                    content="<|17.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51228": AddedToken(
                    content="<|17.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51229": AddedToken(
                    content="<|17.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51230": AddedToken(
                    content="<|17.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51231": AddedToken(
                    content="<|17.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51232": AddedToken(
                    content="<|17.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51233": AddedToken(
                    content="<|17.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51234": AddedToken(
                    content="<|17.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51235": AddedToken(
                    content="<|17.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51236": AddedToken(
                    content="<|17.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51237": AddedToken(
                    content="<|17.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51238": AddedToken(
                    content="<|17.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51239": AddedToken(
                    content="<|17.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51240": AddedToken(
                    content="<|17.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51241": AddedToken(
                    content="<|17.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51242": AddedToken(
                    content="<|17.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51243": AddedToken(
                    content="<|17.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51244": AddedToken(
                    content="<|17.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51245": AddedToken(
                    content="<|17.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51246": AddedToken(
                    content="<|17.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51247": AddedToken(
                    content="<|17.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51248": AddedToken(
                    content="<|17.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51249": AddedToken(
                    content="<|17.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51250": AddedToken(
                    content="<|17.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51251": AddedToken(
                    content="<|17.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51252": AddedToken(
                    content="<|17.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51253": AddedToken(
                    content="<|17.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51254": AddedToken(
                    content="<|17.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51255": AddedToken(
                    content="<|17.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51256": AddedToken(
                    content="<|17.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51257": AddedToken(
                    content="<|17.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51258": AddedToken(
                    content="<|17.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51259": AddedToken(
                    content="<|17.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51260": AddedToken(
                    content="<|17.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51261": AddedToken(
                    content="<|17.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51262": AddedToken(
                    content="<|17.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51263": AddedToken(
                    content="<|17.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51264": AddedToken(
                    content="<|17.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51265": AddedToken(
                    content="<|18.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51266": AddedToken(
                    content="<|18.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51267": AddedToken(
                    content="<|18.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51268": AddedToken(
                    content="<|18.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51269": AddedToken(
                    content="<|18.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51270": AddedToken(
                    content="<|18.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51271": AddedToken(
                    content="<|18.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51272": AddedToken(
                    content="<|18.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51273": AddedToken(
                    content="<|18.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51274": AddedToken(
                    content="<|18.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51275": AddedToken(
                    content="<|18.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51276": AddedToken(
                    content="<|18.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51277": AddedToken(
                    content="<|18.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51278": AddedToken(
                    content="<|18.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51279": AddedToken(
                    content="<|18.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51280": AddedToken(
                    content="<|18.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51281": AddedToken(
                    content="<|18.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51282": AddedToken(
                    content="<|18.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51283": AddedToken(
                    content="<|18.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51284": AddedToken(
                    content="<|18.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51285": AddedToken(
                    content="<|18.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51286": AddedToken(
                    content="<|18.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51287": AddedToken(
                    content="<|18.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51288": AddedToken(
                    content="<|18.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51289": AddedToken(
                    content="<|18.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51290": AddedToken(
                    content="<|18.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51291": AddedToken(
                    content="<|18.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51292": AddedToken(
                    content="<|18.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51293": AddedToken(
                    content="<|18.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51294": AddedToken(
                    content="<|18.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51295": AddedToken(
                    content="<|18.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51296": AddedToken(
                    content="<|18.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51297": AddedToken(
                    content="<|18.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51298": AddedToken(
                    content="<|18.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51299": AddedToken(
                    content="<|18.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51300": AddedToken(
                    content="<|18.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51301": AddedToken(
                    content="<|18.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51302": AddedToken(
                    content="<|18.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51303": AddedToken(
                    content="<|18.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51304": AddedToken(
                    content="<|18.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51305": AddedToken(
                    content="<|18.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51306": AddedToken(
                    content="<|18.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51307": AddedToken(
                    content="<|18.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51308": AddedToken(
                    content="<|18.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51309": AddedToken(
                    content="<|18.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51310": AddedToken(
                    content="<|18.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51311": AddedToken(
                    content="<|18.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51312": AddedToken(
                    content="<|18.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51313": AddedToken(
                    content="<|18.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51314": AddedToken(
                    content="<|18.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51315": AddedToken(
                    content="<|19.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51316": AddedToken(
                    content="<|19.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51317": AddedToken(
                    content="<|19.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51318": AddedToken(
                    content="<|19.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51319": AddedToken(
                    content="<|19.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51320": AddedToken(
                    content="<|19.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51321": AddedToken(
                    content="<|19.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51322": AddedToken(
                    content="<|19.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51323": AddedToken(
                    content="<|19.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51324": AddedToken(
                    content="<|19.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51325": AddedToken(
                    content="<|19.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51326": AddedToken(
                    content="<|19.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51327": AddedToken(
                    content="<|19.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51328": AddedToken(
                    content="<|19.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51329": AddedToken(
                    content="<|19.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51330": AddedToken(
                    content="<|19.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51331": AddedToken(
                    content="<|19.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51332": AddedToken(
                    content="<|19.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51333": AddedToken(
                    content="<|19.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51334": AddedToken(
                    content="<|19.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51335": AddedToken(
                    content="<|19.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51336": AddedToken(
                    content="<|19.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51337": AddedToken(
                    content="<|19.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51338": AddedToken(
                    content="<|19.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51339": AddedToken(
                    content="<|19.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51340": AddedToken(
                    content="<|19.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51341": AddedToken(
                    content="<|19.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51342": AddedToken(
                    content="<|19.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51343": AddedToken(
                    content="<|19.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51344": AddedToken(
                    content="<|19.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51345": AddedToken(
                    content="<|19.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51346": AddedToken(
                    content="<|19.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51347": AddedToken(
                    content="<|19.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51348": AddedToken(
                    content="<|19.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51349": AddedToken(
                    content="<|19.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51350": AddedToken(
                    content="<|19.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51351": AddedToken(
                    content="<|19.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51352": AddedToken(
                    content="<|19.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51353": AddedToken(
                    content="<|19.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51354": AddedToken(
                    content="<|19.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51355": AddedToken(
                    content="<|19.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51356": AddedToken(
                    content="<|19.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51357": AddedToken(
                    content="<|19.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51358": AddedToken(
                    content="<|19.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51359": AddedToken(
                    content="<|19.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51360": AddedToken(
                    content="<|19.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51361": AddedToken(
                    content="<|19.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51362": AddedToken(
                    content="<|19.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51363": AddedToken(
                    content="<|19.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51364": AddedToken(
                    content="<|19.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51365": AddedToken(
                    content="<|20.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51366": AddedToken(
                    content="<|20.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51367": AddedToken(
                    content="<|20.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51368": AddedToken(
                    content="<|20.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51369": AddedToken(
                    content="<|20.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51370": AddedToken(
                    content="<|20.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51371": AddedToken(
                    content="<|20.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51372": AddedToken(
                    content="<|20.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51373": AddedToken(
                    content="<|20.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51374": AddedToken(
                    content="<|20.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51375": AddedToken(
                    content="<|20.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51376": AddedToken(
                    content="<|20.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51377": AddedToken(
                    content="<|20.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51378": AddedToken(
                    content="<|20.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51379": AddedToken(
                    content="<|20.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51380": AddedToken(
                    content="<|20.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51381": AddedToken(
                    content="<|20.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51382": AddedToken(
                    content="<|20.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51383": AddedToken(
                    content="<|20.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51384": AddedToken(
                    content="<|20.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51385": AddedToken(
                    content="<|20.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51386": AddedToken(
                    content="<|20.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51387": AddedToken(
                    content="<|20.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51388": AddedToken(
                    content="<|20.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51389": AddedToken(
                    content="<|20.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51390": AddedToken(
                    content="<|20.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51391": AddedToken(
                    content="<|20.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51392": AddedToken(
                    content="<|20.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51393": AddedToken(
                    content="<|20.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51394": AddedToken(
                    content="<|20.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51395": AddedToken(
                    content="<|20.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51396": AddedToken(
                    content="<|20.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51397": AddedToken(
                    content="<|20.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51398": AddedToken(
                    content="<|20.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51399": AddedToken(
                    content="<|20.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51400": AddedToken(
                    content="<|20.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51401": AddedToken(
                    content="<|20.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51402": AddedToken(
                    content="<|20.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51403": AddedToken(
                    content="<|20.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51404": AddedToken(
                    content="<|20.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51405": AddedToken(
                    content="<|20.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51406": AddedToken(
                    content="<|20.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51407": AddedToken(
                    content="<|20.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51408": AddedToken(
                    content="<|20.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51409": AddedToken(
                    content="<|20.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51410": AddedToken(
                    content="<|20.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51411": AddedToken(
                    content="<|20.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51412": AddedToken(
                    content="<|20.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51413": AddedToken(
                    content="<|20.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51414": AddedToken(
                    content="<|20.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51415": AddedToken(
                    content="<|21.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51416": AddedToken(
                    content="<|21.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51417": AddedToken(
                    content="<|21.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51418": AddedToken(
                    content="<|21.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51419": AddedToken(
                    content="<|21.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51420": AddedToken(
                    content="<|21.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51421": AddedToken(
                    content="<|21.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51422": AddedToken(
                    content="<|21.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51423": AddedToken(
                    content="<|21.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51424": AddedToken(
                    content="<|21.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51425": AddedToken(
                    content="<|21.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51426": AddedToken(
                    content="<|21.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51427": AddedToken(
                    content="<|21.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51428": AddedToken(
                    content="<|21.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51429": AddedToken(
                    content="<|21.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51430": AddedToken(
                    content="<|21.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51431": AddedToken(
                    content="<|21.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51432": AddedToken(
                    content="<|21.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51433": AddedToken(
                    content="<|21.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51434": AddedToken(
                    content="<|21.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51435": AddedToken(
                    content="<|21.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51436": AddedToken(
                    content="<|21.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51437": AddedToken(
                    content="<|21.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51438": AddedToken(
                    content="<|21.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51439": AddedToken(
                    content="<|21.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51440": AddedToken(
                    content="<|21.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51441": AddedToken(
                    content="<|21.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51442": AddedToken(
                    content="<|21.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51443": AddedToken(
                    content="<|21.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51444": AddedToken(
                    content="<|21.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51445": AddedToken(
                    content="<|21.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51446": AddedToken(
                    content="<|21.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51447": AddedToken(
                    content="<|21.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51448": AddedToken(
                    content="<|21.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51449": AddedToken(
                    content="<|21.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51450": AddedToken(
                    content="<|21.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51451": AddedToken(
                    content="<|21.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51452": AddedToken(
                    content="<|21.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51453": AddedToken(
                    content="<|21.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51454": AddedToken(
                    content="<|21.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51455": AddedToken(
                    content="<|21.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51456": AddedToken(
                    content="<|21.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51457": AddedToken(
                    content="<|21.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51458": AddedToken(
                    content="<|21.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51459": AddedToken(
                    content="<|21.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51460": AddedToken(
                    content="<|21.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51461": AddedToken(
                    content="<|21.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51462": AddedToken(
                    content="<|21.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51463": AddedToken(
                    content="<|21.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51464": AddedToken(
                    content="<|21.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51465": AddedToken(
                    content="<|22.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51466": AddedToken(
                    content="<|22.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51467": AddedToken(
                    content="<|22.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51468": AddedToken(
                    content="<|22.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51469": AddedToken(
                    content="<|22.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51470": AddedToken(
                    content="<|22.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51471": AddedToken(
                    content="<|22.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51472": AddedToken(
                    content="<|22.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51473": AddedToken(
                    content="<|22.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51474": AddedToken(
                    content="<|22.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51475": AddedToken(
                    content="<|22.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51476": AddedToken(
                    content="<|22.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51477": AddedToken(
                    content="<|22.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51478": AddedToken(
                    content="<|22.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51479": AddedToken(
                    content="<|22.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51480": AddedToken(
                    content="<|22.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51481": AddedToken(
                    content="<|22.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51482": AddedToken(
                    content="<|22.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51483": AddedToken(
                    content="<|22.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51484": AddedToken(
                    content="<|22.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51485": AddedToken(
                    content="<|22.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51486": AddedToken(
                    content="<|22.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51487": AddedToken(
                    content="<|22.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51488": AddedToken(
                    content="<|22.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51489": AddedToken(
                    content="<|22.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51490": AddedToken(
                    content="<|22.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51491": AddedToken(
                    content="<|22.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51492": AddedToken(
                    content="<|22.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51493": AddedToken(
                    content="<|22.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51494": AddedToken(
                    content="<|22.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51495": AddedToken(
                    content="<|22.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51496": AddedToken(
                    content="<|22.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51497": AddedToken(
                    content="<|22.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51498": AddedToken(
                    content="<|22.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51499": AddedToken(
                    content="<|22.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51500": AddedToken(
                    content="<|22.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51501": AddedToken(
                    content="<|22.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51502": AddedToken(
                    content="<|22.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51503": AddedToken(
                    content="<|22.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51504": AddedToken(
                    content="<|22.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51505": AddedToken(
                    content="<|22.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51506": AddedToken(
                    content="<|22.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51507": AddedToken(
                    content="<|22.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51508": AddedToken(
                    content="<|22.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51509": AddedToken(
                    content="<|22.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51510": AddedToken(
                    content="<|22.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51511": AddedToken(
                    content="<|22.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51512": AddedToken(
                    content="<|22.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51513": AddedToken(
                    content="<|22.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51514": AddedToken(
                    content="<|22.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51515": AddedToken(
                    content="<|23.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51516": AddedToken(
                    content="<|23.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51517": AddedToken(
                    content="<|23.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51518": AddedToken(
                    content="<|23.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51519": AddedToken(
                    content="<|23.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51520": AddedToken(
                    content="<|23.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51521": AddedToken(
                    content="<|23.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51522": AddedToken(
                    content="<|23.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51523": AddedToken(
                    content="<|23.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51524": AddedToken(
                    content="<|23.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51525": AddedToken(
                    content="<|23.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51526": AddedToken(
                    content="<|23.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51527": AddedToken(
                    content="<|23.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51528": AddedToken(
                    content="<|23.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51529": AddedToken(
                    content="<|23.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51530": AddedToken(
                    content="<|23.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51531": AddedToken(
                    content="<|23.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51532": AddedToken(
                    content="<|23.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51533": AddedToken(
                    content="<|23.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51534": AddedToken(
                    content="<|23.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51535": AddedToken(
                    content="<|23.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51536": AddedToken(
                    content="<|23.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51537": AddedToken(
                    content="<|23.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51538": AddedToken(
                    content="<|23.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51539": AddedToken(
                    content="<|23.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51540": AddedToken(
                    content="<|23.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51541": AddedToken(
                    content="<|23.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51542": AddedToken(
                    content="<|23.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51543": AddedToken(
                    content="<|23.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51544": AddedToken(
                    content="<|23.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51545": AddedToken(
                    content="<|23.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51546": AddedToken(
                    content="<|23.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51547": AddedToken(
                    content="<|23.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51548": AddedToken(
                    content="<|23.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51549": AddedToken(
                    content="<|23.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51550": AddedToken(
                    content="<|23.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51551": AddedToken(
                    content="<|23.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51552": AddedToken(
                    content="<|23.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51553": AddedToken(
                    content="<|23.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51554": AddedToken(
                    content="<|23.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51555": AddedToken(
                    content="<|23.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51556": AddedToken(
                    content="<|23.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51557": AddedToken(
                    content="<|23.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51558": AddedToken(
                    content="<|23.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51559": AddedToken(
                    content="<|23.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51560": AddedToken(
                    content="<|23.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51561": AddedToken(
                    content="<|23.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51562": AddedToken(
                    content="<|23.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51563": AddedToken(
                    content="<|23.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51564": AddedToken(
                    content="<|23.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51565": AddedToken(
                    content="<|24.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51566": AddedToken(
                    content="<|24.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51567": AddedToken(
                    content="<|24.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51568": AddedToken(
                    content="<|24.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51569": AddedToken(
                    content="<|24.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51570": AddedToken(
                    content="<|24.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51571": AddedToken(
                    content="<|24.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51572": AddedToken(
                    content="<|24.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51573": AddedToken(
                    content="<|24.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51574": AddedToken(
                    content="<|24.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51575": AddedToken(
                    content="<|24.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51576": AddedToken(
                    content="<|24.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51577": AddedToken(
                    content="<|24.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51578": AddedToken(
                    content="<|24.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51579": AddedToken(
                    content="<|24.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51580": AddedToken(
                    content="<|24.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51581": AddedToken(
                    content="<|24.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51582": AddedToken(
                    content="<|24.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51583": AddedToken(
                    content="<|24.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51584": AddedToken(
                    content="<|24.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51585": AddedToken(
                    content="<|24.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51586": AddedToken(
                    content="<|24.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51587": AddedToken(
                    content="<|24.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51588": AddedToken(
                    content="<|24.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51589": AddedToken(
                    content="<|24.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51590": AddedToken(
                    content="<|24.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51591": AddedToken(
                    content="<|24.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51592": AddedToken(
                    content="<|24.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51593": AddedToken(
                    content="<|24.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51594": AddedToken(
                    content="<|24.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51595": AddedToken(
                    content="<|24.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51596": AddedToken(
                    content="<|24.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51597": AddedToken(
                    content="<|24.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51598": AddedToken(
                    content="<|24.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51599": AddedToken(
                    content="<|24.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51600": AddedToken(
                    content="<|24.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51601": AddedToken(
                    content="<|24.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51602": AddedToken(
                    content="<|24.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51603": AddedToken(
                    content="<|24.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51604": AddedToken(
                    content="<|24.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51605": AddedToken(
                    content="<|24.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51606": AddedToken(
                    content="<|24.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51607": AddedToken(
                    content="<|24.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51608": AddedToken(
                    content="<|24.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51609": AddedToken(
                    content="<|24.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51610": AddedToken(
                    content="<|24.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51611": AddedToken(
                    content="<|24.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51612": AddedToken(
                    content="<|24.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51613": AddedToken(
                    content="<|24.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51614": AddedToken(
                    content="<|24.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51615": AddedToken(
                    content="<|25.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51616": AddedToken(
                    content="<|25.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51617": AddedToken(
                    content="<|25.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51618": AddedToken(
                    content="<|25.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51619": AddedToken(
                    content="<|25.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51620": AddedToken(
                    content="<|25.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51621": AddedToken(
                    content="<|25.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51622": AddedToken(
                    content="<|25.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51623": AddedToken(
                    content="<|25.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51624": AddedToken(
                    content="<|25.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51625": AddedToken(
                    content="<|25.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51626": AddedToken(
                    content="<|25.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51627": AddedToken(
                    content="<|25.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51628": AddedToken(
                    content="<|25.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51629": AddedToken(
                    content="<|25.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51630": AddedToken(
                    content="<|25.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51631": AddedToken(
                    content="<|25.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51632": AddedToken(
                    content="<|25.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51633": AddedToken(
                    content="<|25.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51634": AddedToken(
                    content="<|25.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51635": AddedToken(
                    content="<|25.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51636": AddedToken(
                    content="<|25.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51637": AddedToken(
                    content="<|25.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51638": AddedToken(
                    content="<|25.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51639": AddedToken(
                    content="<|25.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51640": AddedToken(
                    content="<|25.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51641": AddedToken(
                    content="<|25.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51642": AddedToken(
                    content="<|25.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51643": AddedToken(
                    content="<|25.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51644": AddedToken(
                    content="<|25.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51645": AddedToken(
                    content="<|25.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51646": AddedToken(
                    content="<|25.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51647": AddedToken(
                    content="<|25.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51648": AddedToken(
                    content="<|25.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51649": AddedToken(
                    content="<|25.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51650": AddedToken(
                    content="<|25.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51651": AddedToken(
                    content="<|25.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51652": AddedToken(
                    content="<|25.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51653": AddedToken(
                    content="<|25.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51654": AddedToken(
                    content="<|25.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51655": AddedToken(
                    content="<|25.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51656": AddedToken(
                    content="<|25.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51657": AddedToken(
                    content="<|25.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51658": AddedToken(
                    content="<|25.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51659": AddedToken(
                    content="<|25.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51660": AddedToken(
                    content="<|25.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51661": AddedToken(
                    content="<|25.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51662": AddedToken(
                    content="<|25.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51663": AddedToken(
                    content="<|25.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51664": AddedToken(
                    content="<|25.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51665": AddedToken(
                    content="<|26.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51666": AddedToken(
                    content="<|26.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51667": AddedToken(
                    content="<|26.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51668": AddedToken(
                    content="<|26.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51669": AddedToken(
                    content="<|26.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51670": AddedToken(
                    content="<|26.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51671": AddedToken(
                    content="<|26.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51672": AddedToken(
                    content="<|26.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51673": AddedToken(
                    content="<|26.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51674": AddedToken(
                    content="<|26.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51675": AddedToken(
                    content="<|26.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51676": AddedToken(
                    content="<|26.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51677": AddedToken(
                    content="<|26.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51678": AddedToken(
                    content="<|26.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51679": AddedToken(
                    content="<|26.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51680": AddedToken(
                    content="<|26.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51681": AddedToken(
                    content="<|26.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51682": AddedToken(
                    content="<|26.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51683": AddedToken(
                    content="<|26.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51684": AddedToken(
                    content="<|26.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51685": AddedToken(
                    content="<|26.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51686": AddedToken(
                    content="<|26.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51687": AddedToken(
                    content="<|26.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51688": AddedToken(
                    content="<|26.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51689": AddedToken(
                    content="<|26.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51690": AddedToken(
                    content="<|26.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51691": AddedToken(
                    content="<|26.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51692": AddedToken(
                    content="<|26.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51693": AddedToken(
                    content="<|26.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51694": AddedToken(
                    content="<|26.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51695": AddedToken(
                    content="<|26.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51696": AddedToken(
                    content="<|26.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51697": AddedToken(
                    content="<|26.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51698": AddedToken(
                    content="<|26.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51699": AddedToken(
                    content="<|26.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51700": AddedToken(
                    content="<|26.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51701": AddedToken(
                    content="<|26.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51702": AddedToken(
                    content="<|26.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51703": AddedToken(
                    content="<|26.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51704": AddedToken(
                    content="<|26.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51705": AddedToken(
                    content="<|26.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51706": AddedToken(
                    content="<|26.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51707": AddedToken(
                    content="<|26.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51708": AddedToken(
                    content="<|26.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51709": AddedToken(
                    content="<|26.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51710": AddedToken(
                    content="<|26.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51711": AddedToken(
                    content="<|26.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51712": AddedToken(
                    content="<|26.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51713": AddedToken(
                    content="<|26.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51714": AddedToken(
                    content="<|26.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51715": AddedToken(
                    content="<|27.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51716": AddedToken(
                    content="<|27.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51717": AddedToken(
                    content="<|27.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51718": AddedToken(
                    content="<|27.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51719": AddedToken(
                    content="<|27.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51720": AddedToken(
                    content="<|27.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51721": AddedToken(
                    content="<|27.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51722": AddedToken(
                    content="<|27.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51723": AddedToken(
                    content="<|27.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51724": AddedToken(
                    content="<|27.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51725": AddedToken(
                    content="<|27.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51726": AddedToken(
                    content="<|27.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51727": AddedToken(
                    content="<|27.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51728": AddedToken(
                    content="<|27.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51729": AddedToken(
                    content="<|27.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51730": AddedToken(
                    content="<|27.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51731": AddedToken(
                    content="<|27.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51732": AddedToken(
                    content="<|27.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51733": AddedToken(
                    content="<|27.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51734": AddedToken(
                    content="<|27.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51735": AddedToken(
                    content="<|27.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51736": AddedToken(
                    content="<|27.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51737": AddedToken(
                    content="<|27.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51738": AddedToken(
                    content="<|27.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51739": AddedToken(
                    content="<|27.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51740": AddedToken(
                    content="<|27.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51741": AddedToken(
                    content="<|27.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51742": AddedToken(
                    content="<|27.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51743": AddedToken(
                    content="<|27.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51744": AddedToken(
                    content="<|27.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51745": AddedToken(
                    content="<|27.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51746": AddedToken(
                    content="<|27.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51747": AddedToken(
                    content="<|27.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51748": AddedToken(
                    content="<|27.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51749": AddedToken(
                    content="<|27.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51750": AddedToken(
                    content="<|27.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51751": AddedToken(
                    content="<|27.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51752": AddedToken(
                    content="<|27.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51753": AddedToken(
                    content="<|27.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51754": AddedToken(
                    content="<|27.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51755": AddedToken(
                    content="<|27.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51756": AddedToken(
                    content="<|27.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51757": AddedToken(
                    content="<|27.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51758": AddedToken(
                    content="<|27.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51759": AddedToken(
                    content="<|27.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51760": AddedToken(
                    content="<|27.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51761": AddedToken(
                    content="<|27.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51762": AddedToken(
                    content="<|27.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51763": AddedToken(
                    content="<|27.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51764": AddedToken(
                    content="<|27.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51765": AddedToken(
                    content="<|28.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51766": AddedToken(
                    content="<|28.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51767": AddedToken(
                    content="<|28.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51768": AddedToken(
                    content="<|28.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51769": AddedToken(
                    content="<|28.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51770": AddedToken(
                    content="<|28.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51771": AddedToken(
                    content="<|28.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51772": AddedToken(
                    content="<|28.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51773": AddedToken(
                    content="<|28.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51774": AddedToken(
                    content="<|28.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51775": AddedToken(
                    content="<|28.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51776": AddedToken(
                    content="<|28.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51777": AddedToken(
                    content="<|28.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51778": AddedToken(
                    content="<|28.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51779": AddedToken(
                    content="<|28.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51780": AddedToken(
                    content="<|28.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51781": AddedToken(
                    content="<|28.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51782": AddedToken(
                    content="<|28.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51783": AddedToken(
                    content="<|28.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51784": AddedToken(
                    content="<|28.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51785": AddedToken(
                    content="<|28.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51786": AddedToken(
                    content="<|28.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51787": AddedToken(
                    content="<|28.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51788": AddedToken(
                    content="<|28.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51789": AddedToken(
                    content="<|28.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51790": AddedToken(
                    content="<|28.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51791": AddedToken(
                    content="<|28.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51792": AddedToken(
                    content="<|28.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51793": AddedToken(
                    content="<|28.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51794": AddedToken(
                    content="<|28.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51795": AddedToken(
                    content="<|28.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51796": AddedToken(
                    content="<|28.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51797": AddedToken(
                    content="<|28.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51798": AddedToken(
                    content="<|28.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51799": AddedToken(
                    content="<|28.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51800": AddedToken(
                    content="<|28.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51801": AddedToken(
                    content="<|28.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51802": AddedToken(
                    content="<|28.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51803": AddedToken(
                    content="<|28.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51804": AddedToken(
                    content="<|28.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51805": AddedToken(
                    content="<|28.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51806": AddedToken(
                    content="<|28.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51807": AddedToken(
                    content="<|28.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51808": AddedToken(
                    content="<|28.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51809": AddedToken(
                    content="<|28.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51810": AddedToken(
                    content="<|28.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51811": AddedToken(
                    content="<|28.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51812": AddedToken(
                    content="<|28.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51813": AddedToken(
                    content="<|28.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51814": AddedToken(
                    content="<|28.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51815": AddedToken(
                    content="<|29.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51816": AddedToken(
                    content="<|29.02|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51817": AddedToken(
                    content="<|29.04|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51818": AddedToken(
                    content="<|29.06|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51819": AddedToken(
                    content="<|29.08|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51820": AddedToken(
                    content="<|29.10|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51821": AddedToken(
                    content="<|29.12|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51822": AddedToken(
                    content="<|29.14|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51823": AddedToken(
                    content="<|29.16|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51824": AddedToken(
                    content="<|29.18|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51825": AddedToken(
                    content="<|29.20|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51826": AddedToken(
                    content="<|29.22|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51827": AddedToken(
                    content="<|29.24|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51828": AddedToken(
                    content="<|29.26|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51829": AddedToken(
                    content="<|29.28|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51830": AddedToken(
                    content="<|29.30|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51831": AddedToken(
                    content="<|29.32|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51832": AddedToken(
                    content="<|29.34|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51833": AddedToken(
                    content="<|29.36|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51834": AddedToken(
                    content="<|29.38|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51835": AddedToken(
                    content="<|29.40|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51836": AddedToken(
                    content="<|29.42|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51837": AddedToken(
                    content="<|29.44|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51838": AddedToken(
                    content="<|29.46|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51839": AddedToken(
                    content="<|29.48|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51840": AddedToken(
                    content="<|29.50|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51841": AddedToken(
                    content="<|29.52|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51842": AddedToken(
                    content="<|29.54|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51843": AddedToken(
                    content="<|29.56|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51844": AddedToken(
                    content="<|29.58|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51845": AddedToken(
                    content="<|29.60|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51846": AddedToken(
                    content="<|29.62|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51847": AddedToken(
                    content="<|29.64|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51848": AddedToken(
                    content="<|29.66|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51849": AddedToken(
                    content="<|29.68|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51850": AddedToken(
                    content="<|29.70|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51851": AddedToken(
                    content="<|29.72|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51852": AddedToken(
                    content="<|29.74|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51853": AddedToken(
                    content="<|29.76|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51854": AddedToken(
                    content="<|29.78|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51855": AddedToken(
                    content="<|29.80|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51856": AddedToken(
                    content="<|29.82|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51857": AddedToken(
                    content="<|29.84|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51858": AddedToken(
                    content="<|29.86|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51859": AddedToken(
                    content="<|29.88|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51860": AddedToken(
                    content="<|29.90|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51861": AddedToken(
                    content="<|29.92|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51862": AddedToken(
                    content="<|29.94|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51863": AddedToken(
                    content="<|29.96|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51864": AddedToken(
                    content="<|29.98|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
                "51865": AddedToken(
                    content="<|30.00|>",
                    lstrip=False,
                    normalized=True,
                    rstrip=False,
                    single_word=False,
                    special=False,
                ),
            },
            "additional_special_tokens": [
                "<|startoftranscript|>",
                "<|en|>",
                "<|zh|>",
                "<|de|>",
                "<|es|>",
                "<|ru|>",
                "<|ko|>",
                "<|fr|>",
                "<|ja|>",
                "<|pt|>",
                "<|tr|>",
                "<|pl|>",
                "<|ca|>",
                "<|nl|>",
                "<|ar|>",
                "<|sv|>",
                "<|it|>",
                "<|id|>",
                "<|hi|>",
                "<|fi|>",
                "<|vi|>",
                "<|he|>",
                "<|uk|>",
                "<|el|>",
                "<|ms|>",
                "<|cs|>",
                "<|ro|>",
                "<|da|>",
                "<|hu|>",
                "<|ta|>",
                "<|no|>",
                "<|th|>",
                "<|ur|>",
                "<|hr|>",
                "<|bg|>",
                "<|lt|>",
                "<|la|>",
                "<|mi|>",
                "<|ml|>",
                "<|cy|>",
                "<|sk|>",
                "<|te|>",
                "<|fa|>",
                "<|lv|>",
                "<|bn|>",
                "<|sr|>",
                "<|az|>",
                "<|sl|>",
                "<|kn|>",
                "<|et|>",
                "<|mk|>",
                "<|br|>",
                "<|eu|>",
                "<|is|>",
                "<|hy|>",
                "<|ne|>",
                "<|mn|>",
                "<|bs|>",
                "<|kk|>",
                "<|sq|>",
                "<|sw|>",
                "<|gl|>",
                "<|mr|>",
                "<|pa|>",
                "<|si|>",
                "<|km|>",
                "<|sn|>",
                "<|yo|>",
                "<|so|>",
                "<|af|>",
                "<|oc|>",
                "<|ka|>",
                "<|be|>",
                "<|tg|>",
                "<|sd|>",
                "<|gu|>",
                "<|am|>",
                "<|yi|>",
                "<|lo|>",
                "<|uz|>",
                "<|fo|>",
                "<|ht|>",
                "<|ps|>",
                "<|tk|>",
                "<|nn|>",
                "<|mt|>",
                "<|sa|>",
                "<|lb|>",
                "<|my|>",
                "<|bo|>",
                "<|tl|>",
                "<|mg|>",
                "<|as|>",
                "<|tt|>",
                "<|haw|>",
                "<|ln|>",
                "<|ha|>",
                "<|ba|>",
                "<|jw|>",
                "<|su|>",
                "<|yue|>",
                "<|translate|>",
                "<|transcribe|>",
                "<|startoflm|>",
                "<|startofprev|>",
                "<|nospeech|>",
                "<|notimestamps|>",
            ],
            "bos_token": "<|endoftext|>",
            "clean_up_tokenization_spaces": True,
            "eos_token": "<|endoftext|>",
            "errors": "replace",
            "model_max_length": 1000000000000000019884624838656,
            "pad_token": "<|endoftext|>",
            "unk_token": "<|endoftext|>",
        }

class WhisperV1Tokenizer(WhisperV3Tokenizer):
    """
    The tokenizer for Whisper V1 models.
    """
    init_file_urls = {
        "vocab_file": "https://huggingface.co/benjamin-paine/taproot-common/resolve/main/audio-transcription-whisper-tokenizer-vocab.json",
        "merges_file": "https://huggingface.co/benjamin-paine/taproot-common/resolve/main/audio-transcription-whisper-tokenizer-merges.txt",
        "normalizer_file": "https://huggingface.co/benjamin-paine/taproot-common/resolve/main/audio-transcription-whisper-tokenizer-normalizer.json",
        "tokenizer_file": "https://huggingface.co/benjamin-paine/taproot-common/resolve/main/audio-transcription-whisper-tokenizer.json",
    }

    @classmethod
    def get_default_config(cls) -> Optional[Dict[str, Any]]:
        """
        The only modification we HAVE to change is the model max length.
        The other superfluous configuration values will be discarded.
        """
        configuration = super().get_default_config()
        if configuration is not None:
            configuration.update({
                "model_max_length": 1024,
            })
        return configuration
